#include <unistd.h>
#include <string.h>

#include "gstmikmod.h"

extern int need_sync;

static BOOL GST_READER_Eof ( MREADER *reader );
static BOOL GST_READER_Read( MREADER *reader, void *ptr, size_t size );
static int  GST_READER_Get ( MREADER *reader );
static BOOL GST_READER_Seek( MREADER* reader,long offset,int whence );
static long GST_READER_Tell( MREADER* reader );


static BOOL GST_READER_Eof( MREADER *reader )
{
GST_READER *tmp;

  tmp = ( GST_READER * ) reader;

  return tmp->eof;
}


static BOOL GST_READER_Read( MREADER *reader, void *ptr, size_t size )
{
GST_READER *tmp;

  tmp = ( GST_READER * ) reader;
       
  tmp->mik->Buffer = gst_pad_pullregion( tmp->mik->sinkpad, GST_REGION_OFFSET_LEN, tmp->offset, size );

  memcpy( ptr, GST_BUFFER_DATA( tmp->mik->Buffer ), GST_BUFFER_SIZE( tmp->mik->Buffer ));
  tmp->offset = tmp->offset + GST_BUFFER_SIZE( tmp->mik->Buffer );   

  if ( GST_BUFFER_SIZE( tmp->mik->Buffer ) != size )
    tmp->eof = 1;
  else
    tmp->eof = 0;
        
  if (GST_BUFFER_FLAG_IS_SET (tmp->mik->Buffer, GST_BUFFER_FLUSH))        
  	need_sync = 1;
        
  gst_buffer_unref( tmp->mik->Buffer );

  return 1;
}


static int GST_READER_Get ( MREADER *reader )
{
GST_READER *tmp;
int res;

  tmp = ( GST_READER * ) reader;
        
  tmp->mik->Buffer = gst_pad_pullregion( tmp->mik->sinkpad, GST_REGION_OFFSET_LEN, tmp->offset, 1 );
  tmp->offset += GST_BUFFER_SIZE( tmp->mik->Buffer );
  res = *( GST_BUFFER_DATA( tmp->mik->Buffer ));
        
  if ( GST_BUFFER_SIZE( tmp->mik->Buffer ) != 1 )
    tmp->eof = 1;
  else
    tmp->eof = 0;
    
  gst_buffer_unref( tmp->mik->Buffer );

  return res;
}


static BOOL GST_READER_Seek( MREADER *reader, long offset, int whence )
{
GST_READER *tmp;
    
  tmp = ( GST_READER * ) reader;

  if ( whence == SEEK_SET )
    tmp->offset = offset;
  else
    tmp->offset += offset;

  return 1;
}


static long GST_READER_Tell( MREADER *reader )
{
GST_READER *tmp;

  tmp = ( GST_READER * ) reader;

  return tmp->offset;
}


MREADER *GST_READER_new( GstMikMod *mik )
{
GST_READER *reader;
	
  reader = ( GST_READER * ) g_malloc( sizeof( GST_READER ));
  reader->offset = 0;
  reader->eof    = 0;
  reader->mik    = mik;
  if ( reader ) 
  {
     reader->core.Eof  = &GST_READER_Eof;
     reader->core.Read = &GST_READER_Read;
     reader->core.Get  = &GST_READER_Get;
     reader->core.Seek = &GST_READER_Seek;
     reader->core.Tell = &GST_READER_Tell;
  }

  return ( MREADER *)reader;
}


void GST_READER_free ( MREADER *reader )
{
  if ( reader ) 
    g_free( reader );
}
