/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more 
 */

#include <string.h>
#include "gstrtprecv.h"
#include "rtp.h"

/* elementfactory information */
static GstElementDetails gst_rtp_recv_details = {
  "RTP Source",
  "RtpRecv",
  "Streams data via RTP",
  VERSION,
  "Zeeshan Ali <programmer@linuxfreemail.com>",
  "(C) 2001",
};

/* RtpRecv signals and args */
enum {
  SOCKET_CLOSED,
  /* FILL ME */
  LAST_SIGNAL
};

static guint rtprecv_signals[ LAST_SIGNAL ] = { 0 } ;

enum {
  ARG_0,
  ARG_PORT,
  ARG_MTU
};

GST_PADTEMPLATE_FACTORY (src_factory,
	"src",
	GST_PAD_SRC,
	GST_PAD_ALWAYS,
	GST_CAPS_NEW(
	    	"gsm_gsm",
		"audio/x-gsm",
		"rate", GST_PROPS_INT_RANGE (8000, 48000) ),
  	    
	GST_CAPS_NEW(
	    	"mp3",
		"audio/mp3",
		NULL
		//"layer", 	GST_PROPS_INT_RANGE (1, 3),
		//"bitrate", 	GST_PROPS_INT_RANGE (8, 320),
		//"framed", 	GST_PROPS_BOOLEAN (TRUE) 
		),
	
	GST_CAPS_NEW(
	    	"mpeg_video",
		"video/mpeg",
		"mpegversion", 	  GST_PROPS_INT_RANGE (1, 2),
		"systemstream",   GST_PROPS_BOOLEAN (FALSE) ),
		//"sliced", 	  GST_PROPS_BOOLEAN (TRUE) ),
  	    
	GST_CAPS_NEW (
    		"audio_raw",
    		"audio/raw",
    		"format",     GST_PROPS_STRING ("int"),
     	 	"law",        GST_PROPS_INT (0),
       		"endianness", GST_PROPS_INT (G_BYTE_ORDER ),
        	"signed",     GST_PROPS_BOOLEAN (TRUE),
        	"width",      GST_PROPS_INT (16),
      		"depth",      GST_PROPS_INT (16),
      		"rate",       GST_PROPS_INT_RANGE (8000, 48000),
      		"channels",   GST_PROPS_INT (1) ) 
)

static void gst_rtprecv_class_init( GstRtpRecvClass *klass );
static void gst_rtprecv_init( GstRtpRecv *rtprecv );

static GstBuffer* gst_rtprecv_get( GstPad *pad );

static void gst_rtprecv_set_property( GObject *object, guint prop_id, 
		        	      const GValue *value, GParamSpec *pspec );
static void gst_rtprecv_get_property( GObject *object, guint prop_id, 
				      GValue *value, GParamSpec *pspec );

static GstElementStateReturn gst_rtprecv_change_state( GstElement *element );

static void parse_rtcp_app_message( GstPad *pad, Rtcp_APP_message *message ) ;

static GstElementClass *parent_class = NULL;

static GType
gst_rtprecv_get_type(void)
{
  static GType rtprecv_type = 0;

  if (!rtprecv_type) {
    static const GTypeInfo rtprecv_info = {
      sizeof(GstRtpRecvClass),      
      NULL,
      NULL,
      (GClassInitFunc)gst_rtprecv_class_init,
      NULL,
      NULL,
      sizeof(GstRtpRecv),
      0,
      (GInstanceInitFunc)gst_rtprecv_init,
    };
    rtprecv_type = g_type_register_static(GST_TYPE_ELEMENT, "GstRtpRecv", 
		    					&rtprecv_info, 0);
  }
  return rtprecv_type;
}

static void
gst_rtprecv_class_init (GstRtpRecvClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_PORT,
    g_param_spec_int("port","port","port",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); // CHECKME
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_MTU,
    g_param_spec_int("mtu","mtu","mtu",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); // CHECKME

  /* Registering Signals... */
  rtprecv_signals[SOCKET_CLOSED] =
    g_signal_new("socket_closed", G_TYPE_FROM_CLASS( klass ), 
		   G_SIGNAL_RUN_LAST, 
		   G_STRUCT_OFFSET (GstRtpRecvClass, socket_closed ), NULL,
		   NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1,
		   G_TYPE_INT );
  
  gobject_class->set_property = gst_rtprecv_set_property;
  gobject_class->get_property = gst_rtprecv_get_property;

  gstelement_class->change_state = gst_rtprecv_change_state;
  
  klass->socket_closed = NULL ;
}

static void
gst_rtprecv_init (GstRtpRecv *rtprecv)
{
  rtprecv->srcpad = gst_pad_new_from_template( 
		        GST_PADTEMPLATE_GET( src_factory ), "src") ;
  gst_element_add_pad( GST_ELEMENT(rtprecv), rtprecv->srcpad );
  gst_pad_set_get_function( rtprecv->srcpad, 
		            ( GstPadGetFunction )gst_rtprecv_get );

  rtprecv->port = 8000;
  rtprecv->mtu = 160;
}

static GstCaps*
message_to_caps( Rtcp_APP_message* message )
{
  GstCaps* caps ;

  switch( message->payload_type )
  {
    case PAYLOAD_GSM :
      caps = GST_CAPS_NEW (
	       "gsm_gsm",
	       "audio/x-gsm",
	       "rate",        GST_PROPS_INT(( message->frequency ? 
			       		      message->frequency : 8000 ))
      	     );
      break ;
    
    case PAYLOAD_MPA:
      caps = GST_CAPS_NEW (
	       "mp3",
	       "audio/mp3",
	       "layer",  	GST_PROPS_INT( message->layer ?
	       				       message->layer : 3 ),
	       "bitrate",	GST_PROPS_INT( message->frequency ?
		       			       message->frequency : 8 ),
	       "framed",	GST_PROPS_BOOLEAN( message->mpeg_bool2 ) ) ;
      break ;
    
    case PAYLOAD_MPV:
    case PAYLOAD_BMPEG:
      caps = GST_CAPS_NEW (
	       "mpeg_video",
	       "video/mpeg",
	       "mpegversion",  	GST_PROPS_INT( message->layer ?
	       				       message->layer : 2 ),
	       "systemstream",	GST_PROPS_BOOLEAN( message->mpeg_bool1 ), 
	       "sliced",	GST_PROPS_BOOLEAN( message->mpeg_bool2 ) ) ;
	break ;
    
    case PAYLOAD_L16_MONO :
      caps = GST_CAPS_NEW (
	       "gsm_raw", 
	       "audio/raw", 
	       "format",     GST_PROPS_STRING ("int"), /* OK ? */
	       "law", 	     GST_PROPS_INT( 0 ), 
	       "endianness", GST_PROPS_INT( G_BYTE_ORDER ), 
	       "signed",     GST_PROPS_BOOLEAN( TRUE ),
	       "width",      GST_PROPS_INT( 16 ),
	       "depth",      GST_PROPS_INT( 16 ), /* OK ? */
	       "rate",       GST_PROPS_INT(( message->frequency ? 
			       		     message->frequency : 44100 )), 
	       "channels",   GST_PROPS_INT( 1 ) ) ;
      break ;
   
      /* FILL ME  */
  
      default: 
     	/* Unrecognised Type */ 
  	caps = NULL ; 
	//break ;
   }
   
   return caps ; 
}

static void 
parse_rtcp_app_message( GstPad *pad, Rtcp_APP_message* message )
{
  GstRtpRecv *rtprecv = GST_RTPRECV( gst_pad_get_parent( pad ) ) ;

  /* Network to Host Byte ordering */
  message->payload_type = g_ntohl( message->payload_type ) ;
  message->mtu 		= g_ntohl( message->mtu 	 ) ;
  message->frequency    = g_ntohl( message->frequency    ) ;
  message->packets_sent = g_ntohl( message->packets_sent ) ;

  if( message->socket_closed == TRUE )
  {
    /* Emit a message */
    gtk_signal_emit( GTK_OBJECT( rtprecv ), rtprecv_signals[ SOCKET_CLOSED ],
		      message->packets_sent ) ;
  }
 
  else
  { 
    GstCaps *caps ;


    caps = message_to_caps( message ) ;

    if( caps ) {
      gst_pad_set_caps( pad, caps ) ;
    }
    else g_print( "Unknown payload_t %d recieved\n", message->payload_type ) ;
   
    rtprecv->mtu = message->mtu ;
    
    rtprecv->payload_type = message->payload_type ;
    rtprecv->frequency    = message->frequency 	  ;
    rtprecv->layer 	  = message->layer	  ;
    rtprecv->mpeg_bool1   = message->mpeg_bool1   ;
    rtprecv->mpeg_bool2   = message->mpeg_bool2   ;
   
    if( rtprecv->payload_type == PAYLOAD_GSM || 
       rtprecv->payload_type == PAYLOAD_L16_MONO ) 
    rtprecv->frequency = gst_caps_get_int( gst_pad_get_caps( pad ), "rate" ) ;
  }
}  



static GstBuffer* 
gst_rtprecv_get( GstPad *pad )
{
  GstRtpRecv *rtprecv;
  //glong ret;
  GstBuffer *outbuf = NULL ;
  fd_set read_fds;

  g_return_val_if_fail(pad != NULL, NULL );
  g_return_val_if_fail(GST_IS_PAD(pad), NULL );

  rtprecv = GST_RTPRECV (GST_OBJECT_PARENT (pad));

  g_return_val_if_fail (rtprecv != NULL, NULL );
  g_return_val_if_fail (GST_IS_RTPRECV (rtprecv), NULL );

  FD_ZERO (&read_fds);
  FD_SET (rtprecv->conn.rtcp_sock, &read_fds);
  FD_SET (rtprecv->conn.rtp_sock, &read_fds);
  
  if (select (rtprecv->conn.rtcp_sock+1, &read_fds, NULL, NULL, NULL) > 0) {
    if (FD_ISSET (rtprecv->conn.rtcp_sock, &read_fds)) {
	Rtcp_APP_message *message ;
        
	g_print( "Got an RTCP packet\n" ) ;
    	message = rtcp_receive( &rtprecv->conn ) ;
	parse_rtcp_app_message( pad, message ) ;
   	g_free( message ) ;

	outbuf = NULL ; 
    }
    
    else if( FD_ISSET( rtprecv->conn.rtp_sock, &read_fds ) )
    {
      Rtp_Packet packet ;
      Rtcp_APP_message message ;

      packet = rtp_receive( &rtprecv->conn, rtprecv->mtu ) ;

      if( packet == NULL ) {
	g_print( "NULL\n" ) ;
	return NULL ;
      }

      message.payload_type = rtp_packet_get_payload_type( packet ) ;

      if( message.payload_type != rtprecv->payload_type )
      {
        GstCaps *caps ;
	
   	message.frequency  = 0 ;
   	message.mtu = rtprecv->mtu ;
   	message.layer	    = 0 ;
   	message.mpeg_bool1 = FALSE ;
   	message.mpeg_bool2 = TRUE  ;
	
    	g_print( "new payload_t %d recieved\n", message.payload_type ) ;
	caps = message_to_caps( &message ) ;
	if( caps ) {
	   gst_pad_set_caps( rtprecv->srcpad, caps ) ;
	}
    	else g_print( "Unknown payload_t %d recieved\n", 
			message.payload_type ) ;
	
	rtprecv->payload_type = message.payload_type ;
      }

      outbuf = gst_buffer_new() ;
      GST_BUFFER_SIZE( outbuf ) = rtp_packet_get_payload_len( packet ) ;
      GST_BUFFER_DATA( outbuf ) = g_malloc( GST_BUFFER_SIZE( outbuf ) ) ;
   
      memcpy( GST_BUFFER_DATA( outbuf ), rtp_packet_get_payload( packet ),
	        GST_BUFFER_SIZE( outbuf ) ) ;
      rtp_packet_free( packet ) ;
    }
  }
  else {
    perror ("select");
  }
  
  return outbuf;
}

static void
gst_rtprecv_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GstRtpRecv *rtprecv;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPRECV(object));
  rtprecv = GST_RTPRECV(object);

  switch (prop_id) {
  case ARG_PORT:
    rtprecv->port = g_value_get_int (value);
    break;
  case ARG_MTU:
    rtprecv->mtu = g_value_get_int (value);
  default:
    break;
  }
}

static void
gst_rtprecv_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GstRtpRecv *rtprecv;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPRECV(object));
  rtprecv = GST_RTPRECV(object);

  switch (prop_id) {
  case ARG_PORT:
    g_value_set_int (value, rtprecv->port);
    break;
  case ARG_MTU:
    g_value_set_int (value, rtprecv->mtu);
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static GstElementStateReturn
gst_rtprecv_change_state (GstElement *element)
{
  GstRtpRecv *rtprecv;

  g_return_val_if_fail (GST_IS_RTPRECV (element), GST_STATE_FAILURE);

  rtprecv = GST_RTPRECV(element);

  GST_DEBUG (0,"state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  switch (GST_STATE_TRANSITION (element)) {
   
  case GST_STATE_NULL_TO_READY:

     if (rtp_server_init( &rtprecv->conn, rtprecv->port )) {
       g_print("initialised RTP successfully...\n");
     }
     else {
       g_warning ("couldn't initialise RTP...\n");
       return GST_STATE_FAILURE;
     }
     
     rtp_server_connection( &rtprecv->conn );
     break;
   
   case GST_STATE_READY_TO_NULL:
     /*shout_disconnect (&icecastrecv->conn);*/
    close( rtprecv->conn.rtp_sock ) ; /* wtay: OK ? */
    close( rtprecv->conn.rtcp_sock ) ;
    rtprecv->conn.rtp_sock = 0 ;
    rtprecv->conn.rtcp_sock = 0 ;
    break;
   
   default:
     break;
  }

  /* if we haven't failed already, give the parent class a chance to ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

gboolean
gst_rtprecv_plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *recv ;

  recv = gst_elementfactory_new ("rtprecv", 
		       GST_TYPE_RTPRECV, &gst_rtp_recv_details);
  g_return_val_if_fail (recv != NULL, FALSE);

  gst_elementfactory_add_padtemplate (recv, GST_PADTEMPLATE_GET (src_factory));

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (recv) );

  return TRUE;
}
