/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include "gstrtpsend.h"
#include "rtp.h"

/* elementfactory information */
static GstElementDetails gst_rtp_send_details = {
  "RTP Sink",
  "RtpSend",
  "Streams data via RTP",
  VERSION,
  "Zaheer Merali <zaheer@bellworldwide.net>, " 
  "Zeeshan Ali <programmer@linuxfreemail.com",
  "(C) 2001",
};

/* RtpSend signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_IP_LIST,
  ARG_IP,
  ARG_PORT,
  ARG_PAYLOAD_TYPE,
  ARG_MTU
};

GST_PADTEMPLATE_FACTORY (sink_factory,
	"sink",
 	GST_PAD_SINK,
	GST_PAD_ALWAYS,
	GST_CAPS_NEW(
    		"gsm_gsm",
    		"audio/x-gsm",
      		"rate", GST_PROPS_INT_RANGE (8000, 48000) ),
	
	GST_CAPS_NEW(
    		"mp3",
    		"audio/mp3",
		"layer", 	GST_PROPS_INT_RANGE (1, 3),
		"bitrate", 	GST_PROPS_INT_RANGE (8, 320),
		"framed", 	GST_PROPS_BOOLEAN (TRUE) ),
	
	
	GST_CAPS_NEW (
    		"mpeg_video",
    		"video/mpeg",
		NULL
      		//"mpegversion",      GST_PROPS_INT_RANGE (1, 2),
        	//"systemstream",     GST_PROPS_BOOLEAN (FALSE),
        	//"sliced",     	    GST_PROPS_BOOLEAN (TRUE) ),
		),
	
	GST_CAPS_NEW (
    		"audio_raw",
    		"audio/raw",
      		"format",     	GST_PROPS_STRING ("int"),
        	"law",        	GST_PROPS_INT (0),
        	"endianness", 	GST_PROPS_INT (G_BYTE_ORDER ),
        	"signed",     	GST_PROPS_BOOLEAN (TRUE),
        	"width",      	GST_PROPS_INT (16),
      		"depth",      	GST_PROPS_INT (16),
      		"rate",       	GST_PROPS_INT_RANGE (8000, 48000),
      		"channels",   	GST_PROPS_INT (1) ) 
)

static void			gst_rtpsend_class_init		(GstRtpSendClass *klass);
static void			gst_rtpsend_init		(GstRtpSend *rtpsend);

static void			gst_rtpsend_chain		(GstPad *pad, GstBuffer *buf);

static void			gst_rtpsend_set_property	(GObject *object, guint prop_id, 
								 const GValue *value, GParamSpec *pspec);
static void			gst_rtpsend_get_property	(GObject *object, guint prop_id, 
								 GValue *value, GParamSpec *pspec);

static void 			gst_rtpsend_newcaps		(GstPad *pad, GstCaps *caps);
static GstElementStateReturn	gst_rtpsend_change_state	(GstElement *element);

static GstElementClass *parent_class = NULL;
//static guint gst_icecastsend_signals[LAST_SIGNAL] = { 0 };

static GType
gst_rtpsend_get_type (void)
{
  static GType rtpsend_type = 0;

  if (!rtpsend_type) {
    static const GTypeInfo rtpsend_info = {
      sizeof(GstRtpSendClass),      
      NULL,
      NULL,
      (GClassInitFunc)gst_rtpsend_class_init,
      NULL,
      NULL,
      sizeof(GstRtpSend),
      0,
      (GInstanceInitFunc)gst_rtpsend_init,
    };
    rtpsend_type = g_type_register_static(GST_TYPE_ELEMENT, "GstRtpSend", &rtpsend_info, 0);
  }
  return rtpsend_type;
}

static void
gst_rtpsend_class_init (GstRtpSendClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_IP_LIST,
    g_param_spec_pointer("ip_list","ip_list","ip_list",
                        G_PARAM_READWRITE)); // CHECKME
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_IP,
    g_param_spec_string("ip","ip","ip",
                        NULL, G_PARAM_READWRITE)); // CHECKME
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_PORT,
    g_param_spec_int("port","port","port",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); // CHECKME
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_PAYLOAD_TYPE,
    g_param_spec_string("payload_type","payload_type","payload_type",
                        NULL, G_PARAM_READWRITE)); // CHECKME
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_MTU,
    g_param_spec_int("mtu","mtu","mtu",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); // CHECKME

  gobject_class->set_property = gst_rtpsend_set_property;
  gobject_class->get_property = gst_rtpsend_get_property;

  gstelement_class->change_state = gst_rtpsend_change_state;
}

static void
gst_rtpsend_init (GstRtpSend *rtpsend)
{
  rtpsend->sinkpad = gst_pad_new_from_template( 
		         GST_PADTEMPLATE_GET( sink_factory ), "sink") ;
  gst_element_add_pad(GST_ELEMENT(rtpsend),rtpsend->sinkpad);
  gst_pad_set_chain_function(rtpsend->sinkpad,gst_rtpsend_chain);
  gst_pad_set_newcaps_function(rtpsend->sinkpad,gst_rtpsend_newcaps);

  rtpsend->payload_type = PAYLOAD_GSM ;
  
  rtpsend->frequency = 8000 ;	
  rtpsend->layer     = 3    ;	
  rtpsend->mpeg_bool1 = FALSE ;	
  rtpsend->mpeg_bool2 = TRUE  ;	
  
  rtpsend->ip = NULL; // for append func. to work
  rtpsend->ip = g_slist_append( rtpsend->ip, "127.0.0.1" );
  rtpsend->port = 8000 ;
  rtpsend->mtu = 160 ;
  
  rtpsend->conn = NULL ;
  
  rtpsend->clock = gst_clock_get_system();
  gst_clock_register( rtpsend->clock, GST_OBJECT( rtpsend ) ) ;
}

static void 
gst_rtpsend_newcaps( GstPad *pad, GstCaps *caps ) 
{
  GstRtpSend *rtpsend ;
  const gchar *mime = gst_caps_get_mime( caps ) ;

  g_print( "new caps called with caps' mime type: %s\n", mime ) ;
  rtpsend = GST_RTPSEND( gst_pad_get_parent( pad ) ) ;
 
  if( strcmp( mime, "audio/x-gsm" ) == 0 ) {
    rtpsend->payload_type = PAYLOAD_GSM ;
    rtpsend->frequency = gst_caps_get_int( caps, "rate" ) ;
    rtpsend->mtu = 160 ;
  }

  else if( strcmp( mime, "audio/mp3" ) == 0 ) {
    rtpsend->payload_type = PAYLOAD_MPA ;
    rtpsend->mtu = 512 ; 
    rtpsend->frequency 	  = gst_caps_get_int( caps, "bitrate" ) ;
    rtpsend->layer 	  = gst_caps_get_int( caps, "layer" ) ;
    rtpsend->mpeg_bool2   = gst_caps_get_boolean( caps, "framed" ) ;
  }
  
  else if( strcmp( mime, "video/mpeg" ) == 0 ) {
    gboolean bool ;

    if( ( bool = gst_caps_get_boolean( caps, "systemstream" ) ) )
    rtpsend->payload_type = PAYLOAD_BMPEG ;
    else rtpsend->payload_type = PAYLOAD_MPV ;
    
    rtpsend->mtu = 1024 ;
    rtpsend->layer 	  = gst_caps_get_int( caps, "mpegversion" ) ;
    rtpsend->mpeg_bool1   = bool ;
    rtpsend->mpeg_bool2   = gst_caps_get_boolean( caps, "sliced" ) ;
  }
  
  else if( strcmp( mime, "audio/raw" ) == 0 ) { 
   if( gst_caps_get_int( caps, "law" ) == 0 ) 
    if( gst_caps_get_int( caps, "width" ) == 16 ) 
     if( gst_caps_get_boolean( caps, "channels" ) == 1 ) 
      if( gst_caps_get_boolean( caps, "signed" ) == TRUE ) {
 	rtpsend->payload_type = PAYLOAD_L16_MONO ; 
 	rtpsend->mtu = 1024 ; 
    	rtpsend->frequency = gst_caps_get_int( caps, "rate" ) ;
      }
  }
  
  /* FILL ME  */
 
  /* Unrecognised Type */ 
  else  {
    rtpsend->payload_type = 0 ;
    g_print( "Unsupported media type\n" ) ;
  }
  
  if( rtpsend->conn != NULL ) { 
     Rtcp_APP_message message ;
     
     message.socket_closed = FALSE ;
     message.payload_type  = g_htonl( rtpsend->payload_type );
     message.mtu 	   = g_htonl( rtpsend->mtu );
     message.frequency 	   = g_htonl( rtpsend->frequency );
     message.layer 	   = rtpsend->layer ;
     message.mpeg_bool1    = rtpsend->mpeg_bool1 ;
     message.mpeg_bool2    = rtpsend->mpeg_bool2 ;
     
     g_print( "Trying to send Rtcp compound\n" ) ;
     rtcp_send( rtpsend->conn, message );
  }
}


static void
gst_rtpsend_chain (GstPad *pad, GstBuffer *buf)
{
  GstRtpSend *rtpsend;
  GstClockTimeDiff jitter;
  //glong ret;
  guint32 i;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);

  rtpsend = GST_RTPSEND (GST_OBJECT_PARENT (pad));

  g_return_if_fail (rtpsend != NULL);
  g_return_if_fail (GST_IS_RTPSEND (rtpsend));

  if (!GST_BUFFER_FLAG_IS_SET(buf, GST_BUFFER_FLUSH)) {
    GST_DEBUG (0,"rtpsend: clock wait: %llu\n", GST_BUFFER_TIMESTAMP(buf));

    jitter = gst_clock_current_diff(rtpsend->clock, GST_BUFFER_TIMESTAMP (buf));

    if (jitter > 500000 || jitter < -500000)
    {
      GST_DEBUG( 0, "jitter: %lld\n", jitter);
      gst_clock_set( rtpsend->clock, GST_BUFFER_TIMESTAMP ( buf ) );
    }
    else {
      gst_clock_wait( rtpsend->clock, GST_BUFFER_TIMESTAMP( buf ), 
		      GST_OBJECT( rtpsend ) );
    }
  }
  
  for(i=0;i<GST_BUFFER_SIZE(buf);i+=rtpsend->mtu) {
    if (GST_BUFFER_SIZE(buf)-i>rtpsend->mtu) {
      rtp_send (rtpsend->conn, GST_BUFFER_DATA (buf)+i,
		rtpsend->mtu, rtpsend->payload_type, 20);
    }
    else {
      rtp_send (rtpsend->conn, GST_BUFFER_DATA (buf)+i,
		GST_BUFFER_SIZE (buf) - i, rtpsend->payload_type, 20);
    }
  }
  /*
  if (!ret) {
    g_warning ("rtp send error...\n");
  }
  */
  
  gst_buffer_unref (buf);
}

static void
gst_rtpsend_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GstRtpSend *rtpsend;
  gchar *payload_type_str ;
  GSList *iterator = NULL;
  guint8 ret_val = 0;
  gchar **ip, *address;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPSEND(object));
  rtpsend = GST_RTPSEND(object);

  switch (prop_id) {
  case ARG_IP_LIST:
     rtpsend->ip = g_value_get_pointer( value );
     
     rtp_close_all( &rtpsend->conn );
     rtp_client_connection (&rtpsend->conn);

     for( iterator = rtpsend->ip; 
		     	iterator; iterator = g_slist_next( iterator ) ) {
	    address = (gchar *)iterator->data;
	    ret_val = ret_val | rtp_connection_call(&rtpsend->conn,
						    address,rtpsend->port);
     }
	
     if( ret_val ) {
	g_print("new IP list launched successfully...\n");
     }
     break;
  case ARG_IP:
     if (rtpsend->ip != NULL) {
  	rtpsend->ip = g_slist_remove( rtpsend->ip, "127.0.0.1" ) ;
     }
     
     rtp_client_connection (&rtpsend->conn);
     for( ip = g_strsplit( g_value_get_string( value ), ",", 0 );
     	  *ip; ip++ ) {
          rtpsend->ip = g_slist_append( rtpsend->ip, *ip );
	  ret_val = ret_val | rtp_connection_call(&rtpsend->conn,
					     *ip,rtpsend->port);
     }
     
     if( ret_val ) {
	g_print("IPs added successfully...\n");
     }
     break; 
  case ARG_PORT:
    rtpsend->port = g_value_get_int (value);
    break;
  case ARG_PAYLOAD_TYPE :
    payload_type_str = g_strdup (g_value_get_string (value));
  
    if( strcmp( payload_type_str, "gsm" ) == 0 )
    	rtpsend->payload_type = PAYLOAD_GSM ;
    
    if( strcmp( payload_type_str, "mp1" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_MPA ;
    	rtpsend->layer = 1 ;
    	rtpsend->mtu = 512 ;
    }

    if( strcmp( payload_type_str, "mp2" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_MPA ;
    	rtpsend->layer = 2 ;
    	rtpsend->mtu = 512 ;
    }

    if( strcmp( payload_type_str, "mp3" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_MPA ;
    	rtpsend->layer = 3 ;
    	rtpsend->mtu = 512 ;
    }

    if( strcmp( payload_type_str, "mpeg1" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_MPV ;
    	rtpsend->layer = 1 ;
    	rtpsend->mpeg_bool1   = FALSE ;
    	rtpsend->mtu = 1024 ;
    }

    if( strcmp( payload_type_str, "mpeg2" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_MPV ;
    	rtpsend->layer = 2 ;
    	rtpsend->mpeg_bool1   = FALSE ;
    	rtpsend->mtu = 1024 ;
    }

    if( strcmp( payload_type_str, "mpeg1_sys" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_BMPEG ;
    	rtpsend->layer = 1 ;
    	rtpsend->mpeg_bool1 = TRUE ;
    	rtpsend->mtu = 1024 ;
    }
    
    if( strcmp( payload_type_str, "mpeg2_sys" ) == 0 ) {
    	rtpsend->payload_type = PAYLOAD_BMPEG ;
    	rtpsend->layer = 2 ;
    	rtpsend->mpeg_bool1 = TRUE ;
    	rtpsend->mtu = 1024 ;
    } 
    
    if( strcmp( payload_type_str, "raw_audio" ) == 0 )
    	rtpsend->payload_type = PAYLOAD_L16_MONO ;
  
    break;
  case ARG_MTU:
    rtpsend->mtu = g_value_get_int (value);
  default:
    break;
  }
}

static void
gst_rtpsend_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GstRtpSend *rtpsend;
  GSList *iterator;
  gchar *ip1 = NULL, *ip2 = NULL;
  
  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_RTPSEND(object));
  rtpsend = GST_RTPSEND(object);

  switch (prop_id) {
  case ARG_IP_LIST:
    g_value_set_pointer( value, rtpsend->ip );
    break;
  case ARG_IP:
    for( iterator = rtpsend->ip ; iterator ; 
		    iterator = g_slist_next( iterator ) ) {
	if( ip1 == NULL ) {
	    ip1 = (gchar *)iterator->data;
	}
	else {
   	    /* This code is'nt as silly as it seems: */
	    g_free( ip2 );
            ip2 = ip1;	
	    ip1 = g_strjoin( ",", ip1, (gchar *)iterator->data, NULL );
	}
    }
    g_value_set_string( value, ip1 );
    break;
  case ARG_PORT:
    g_value_set_int (value, rtpsend->port);
    break;
  case ARG_PAYLOAD_TYPE:
    if( rtpsend->payload_type == PAYLOAD_GSM )
       g_value_set_string (value, "gsm" );
    
    if( rtpsend->payload_type == PAYLOAD_MPA ) {
       if( rtpsend->layer == 1 ) 
          g_value_set_string (value, "mp1" );
       if( rtpsend->layer == 2 ) 
          g_value_set_string (value, "mp2" );
       if( rtpsend->layer == 3 ) 
          g_value_set_string (value, "mp3" );
    }
    
    if( rtpsend->payload_type == PAYLOAD_MPV ) {
       if( rtpsend->layer == 1 )
       	  g_value_set_string (value, "mpeg1" );
       if( rtpsend->layer == 2 )
       	  g_value_set_string (value, "mpeg2" );
    }
    
    if( rtpsend->payload_type == PAYLOAD_BMPEG ) {
       if( rtpsend->layer == 1 )
       	  g_value_set_string (value, "mpeg1_sys" );
       if( rtpsend->layer == 2 )
       	  g_value_set_string (value, "mpeg2_sys" );
    }
    
    if( rtpsend->payload_type == PAYLOAD_L16_MONO )
    g_value_set_string (value, "raw_audio" );
    
    break;
  case ARG_MTU:
    g_value_set_int (value, rtpsend->mtu);
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static GstElementStateReturn
gst_rtpsend_change_state (GstElement *element)
{
  GstRtpSend *rtpsend;
  Rtcp_APP_message message ;
  GSList *iterator = NULL;
  gchar *ip;
  guint8 ret_val = 0;

  g_return_val_if_fail (GST_IS_RTPSEND (element), GST_STATE_FAILURE);

  rtpsend = GST_RTPSEND(element);

  GST_DEBUG (0,"state pending %d\n", GST_STATE_PENDING (element));

  /* if going down into NULL state, close the file if it's open */
  switch (GST_STATE_TRANSITION (element)) {
   case GST_STATE_NULL_TO_READY:
	rtp_client_connection (&rtpsend->conn);

 	for( iterator = rtpsend->ip; 
		     	iterator; iterator = g_slist_next( iterator ) ) {
	    ip = (gchar *)iterator->data;
	    ret_val = ret_val | rtp_connection_call(&rtpsend->conn,ip,
							rtpsend->port);
	}
	
        if( ret_val ) {
	   g_print("initialised RTP successfully...\n");
  	        
	   message.socket_closed = FALSE ;
	   message.payload_type  = g_htonl( rtpsend->payload_type );
	   message.mtu 	      = g_htonl( rtpsend->mtu );
	   message.frequency     = g_htonl( rtpsend->frequency );
	   message.layer 	     = rtpsend->layer ;
	   message.mpeg_bool1    = rtpsend->mpeg_bool1 ;
	   message.mpeg_bool2    = rtpsend->mpeg_bool2 ;
       
	    g_print( "Trying to send Rtcp packet\n" ) ;
	    rtcp_send( rtpsend->conn, message ) ;
        }
        else {
	   g_warning ("couldn't initialise RTP...\n");
	   return GST_STATE_FAILURE;
        }
        break;
  
   case GST_STATE_READY_TO_NULL:
     /*shout_disconnect (&icecastsend->conn);*/
    message.socket_closed = TRUE ;
    rtcp_send( rtpsend->conn, message ) ;
    rtp_close_all( &rtpsend->conn );
    break;
   
   default:
     break;
  }

  /* if we haven't failed already, give the parent class a chance to ;-) */
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

gboolean
gst_rtpsend_plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *send ;

  send = gst_elementfactory_new ("rtpsend",
                           GST_TYPE_RTPSEND, &gst_rtp_send_details);
  g_return_val_if_fail (send != NULL, FALSE);

  gst_elementfactory_add_padtemplate (send, GST_PADTEMPLATE_GET (sink_factory));

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (send) );

  return TRUE;
}

